@@ -22,6 +22,60 @@ module ActiveRecord::Import #:nodoc:
       super "Missing column for value <#{name}> at index #{index}"
     end
   end
+
+  class Validator
+    def initialize(options = {})
+      @options = options
+    end
+
+    def valid_model?(model)
+      validation_context = @options[:validate_with_context]
+      validation_context ||= (model.new_record? ? :create : :update)
+
+      current_context = model.send(:validation_context)
+      begin
+        model.send(:validation_context=, validation_context)
+        model.errors.clear
+
+        validate_callbacks = model._validate_callbacks.dup
+        validate_callbacks.each do |callback|
+          validate_callbacks.delete(callback) if callback.raw_filter.is_a? ActiveRecord::Validations::UniquenessValidator
+        end
+
+        model.run_callbacks(:validation) do
+          if defined?(ActiveSupport::Callbacks::Filters::Environment) # ActiveRecord >= 4.1
+            runner = validate_callbacks.compile
+            env = ActiveSupport::Callbacks::Filters::Environment.new(model, false, nil)
+            if runner.respond_to?(:call) # ActiveRecord < 5.1
+              runner.call(env)
+            else # ActiveRecord 5.1
+              # Note that this is a gross simplification of ActiveSupport::Callbacks#run_callbacks.
+              # It's technically possible for there to exist an "around" callback in the
+              # :validate chain, but this would be an aberration, since Rails doesn't define
+              # "around_validate". Still, rather than silently ignoring such callbacks, we
+              # explicitly raise a RuntimeError, since activerecord-import was asked to perform
+              # validations and it's unable to do so.
+              #
+              # The alternative here would be to copy-and-paste the bulk of the
+              # ActiveSupport::Callbacks#run_callbacks method, which is undesirable if there's
+              # no real-world use case for it.
+              raise "The :validate callback chain contains an 'around' callback, which is unsupported" unless runner.final?
+              runner.invoke_before(env)
+              runner.invoke_after(env)
+            end
+          elsif validate_callbacks.method(:compile).arity == 0 # ActiveRecord = 4.0
+            model.instance_eval validate_callbacks.compile
+          else # ActiveRecord 3.x
+            model.instance_eval validate_callbacks.compile(nil, model)
+          end
+        end
+
+        model.errors.empty?
+      ensure
+        model.send(:validation_context=, current_context)
+      end
+    end
+  end
 end
 
 class ActiveRecord::Associations::CollectionProxy
@@ -97,19 +151,6 @@ end
 
 class ActiveRecord::Base
   class << self
-    # use tz as set in ActiveRecord::Base
-    tproc = lambda do
-      ActiveRecord::Base.default_timezone == :utc ? Time.now.utc : Time.now
-    end
-
-    AREXT_RAILS_COLUMNS = {
-      create: { "created_on" => tproc,
-                "created_at" => tproc },
-      update: { "updated_on" => tproc,
-                "updated_at" => tproc }
-    }.freeze
-    AREXT_RAILS_COLUMN_NAMES = AREXT_RAILS_COLUMNS[:create].keys + AREXT_RAILS_COLUMNS[:update].keys
-
     # Returns true if the current database connection adapter
     # supports import functionality, otherwise returns false.
     def supports_import?(*args)
@@ -144,6 +185,9 @@ class ActiveRecord::Base
     #
     # == Usage
     #  Model.import array_of_models
+    #  Model.import column_names, array_of_models
+    #  Model.import array_of_hash_objects
+    #  Model.import column_names, array_of_hash_objects
     #  Model.import column_names, array_of_values
     #  Model.import column_names, array_of_values, options
     #
@@ -171,11 +215,14 @@ class ActiveRecord::Base
     # == Options
     # * +validate+ - true|false, tells import whether or not to use
     #    ActiveRecord validations. Validations are enforced by default.
-    # * +ignore+ - true|false, tells import to use MySQL's INSERT IGNORE
-    #    to discard records that contain duplicate keys.
-    # * +on_duplicate_key_ignore+ - true|false, tells import to use
-    #    Postgres 9.5+ ON CONFLICT DO NOTHING. Cannot be enabled on a
-    #    recursive import.
+    # * +ignore+ - true|false, an alias for on_duplicate_key_ignore.
+    # * +on_duplicate_key_ignore+ - true|false, tells import to discard
+    #    records that contain duplicate keys. For Postgres 9.5+ it adds
+    #    ON CONFLICT DO NOTHING, for MySQL it uses INSERT IGNORE, and for
+    #    SQLite it uses INSERT OR IGNORE. Cannot be enabled on a
+    #    recursive import. For database adapters that normally support
+    #    setting primary keys on imported objects, this option prevents
+    #    that from occurring.
     # * +on_duplicate_key_update+ - an Array or Hash, tells import to
     #    use MySQL's ON DUPLICATE KEY UPDATE or Postgres 9.5+ ON CONFLICT
     #    DO UPDATE ability. See On Duplicate Key Update below.
@@ -186,7 +233,7 @@ class ActiveRecord::Base
     #   (if false) even if record timestamps is disabled in ActiveRecord::Base
     # * +recursive+ - true|false, tells import to import all has_many/has_one
     #   associations if the adapter supports setting the primary keys of the
-    #   newly imported objects.
+    #   newly imported objects. PostgreSQL only.
     # * +batch_size+ - an integer value to specify the max number of records to
     #   include per insert. Defaults to the total number of records to import.
     #
@@ -199,6 +246,15 @@ class ActiveRecord::Base
     #            BlogPost.new author_name: 'Zach Dennis', title: 'AREXT3' ]
     #  BlogPost.import posts
     #
+    #  # Example using array_of_hash_objects
+    #  values = [ {author_name: 'zdennis', title: 'test post'} ], [ {author_name: 'jdoe', title: 'another test post'} ] ]
+    #  BlogPost.import values
+    #
+    #  # Example using column_names and array_of_hash_objects
+    #  columns = [ :author_name, :title ]
+    #  values = [ {author_name: 'zdennis', title: 'test post'} ], [ {author_name: 'jdoe', title: 'another test post'} ] ]
+    #  BlogPost.import columns, values
+    #
     #  # Example using column_names and array_of_values
     #  columns = [ :author_name, :title ]
     #  values = [ [ 'zdennis', 'test post' ], [ 'jdoe', 'another test post' ] ]
@@ -246,7 +302,8 @@ class ActiveRecord::Base
     # == On Duplicate Key Update (Postgres 9.5+)
     #
     # The :on_duplicate_key_update option can be an Array or a Hash with up to
-    # two attributes, :conflict_target or :constraint_name and :columns.
+    # three attributes, :conflict_target (and optionally :index_predicate) or
+    # :constraint_name, and :columns.
     #
     # ==== Using an Array
     #
@@ -260,10 +317,11 @@ class ActiveRecord::Base
     #
     # ====  Using a Hash
     #
-    # The :on_duplicate_update option can be a hash with up to two attributes,
-    # :conflict_target or constraint_name, and :columns. Unlike MySQL, Postgres
-    # requires the conflicting constraint to be explicitly specified. Using this
-    # option allows you to specify a constraint other than the primary key.
+    # The :on_duplicate_key_update option can be a hash with up to three
+    # attributes, :conflict_target (and optionally :index_predicate) or
+    # :constraint_name, and :columns. Unlike MySQL, Postgres requires the
+    # conflicting constraint to be explicitly specified. Using this option
+    # allows you to specify a constraint other than the primary key.
     #
     # ====== :conflict_target
     #
@@ -273,7 +331,16 @@ class ActiveRecord::Base
     # but it is the preferred method of identifying a constraint. It will
     # default to the primary key. Below is an example:
     #
-    #   BlogPost.import columns, values, on_duplicate_key_update: { conflict_target: [:author_id, :slug], columns: [ :date_modified ] }
+    #   BlogPost.import columns, values, on_duplicate_key_update: { conflict_target: [ :author_id, :slug ], columns: [ :date_modified ] }
+    #
+    # ====== :index_predicate
+    #
+    # The :index_predicate attribute optionally specifies a WHERE condition
+    # on :conflict_target, which is required for matching against partial
+    # indexes. This attribute is ignored if :constraint_name is included.
+    # Below is an example:
+    #
+    #   BlogPost.import columns, values, on_duplicate_key_update: { conflict_target: [ :author_id, :slug ], index_predicate: 'status <> 0', columns: [ :date_modified ] }
     #
     # ====== :constraint_name
     #
@@ -283,6 +350,15 @@ class ActiveRecord::Base
     #
     #   BlogPost.import columns, values, on_duplicate_key_update: { constraint_name: :blog_posts_pkey, columns: [ :date_modified ] }
     #
+    # ====== :condition
+    #
+    # The :condition attribute optionally specifies a WHERE condition
+    # on :conflict_action. Only rows for which this expression returns true will be updated.
+    # Note that it's evaluated last, after a conflict has been identified as a candidate to update.
+    # Below is an example:
+    #
+    #   BlogPost.import columns, values, on_duplicate_key_update: { conflict_target: [ :author_id ], condition: "blog_posts.title NOT LIKE '%sample%'", columns: [ :author_name ] }
+
     # ====== :columns
     #
     # The :columns attribute can be either an Array or a Hash.
@@ -307,7 +383,7 @@ class ActiveRecord::Base
     # This returns an object which responds to +failed_instances+ and +num_inserts+.
     # * failed_instances - an array of objects that fails validation and were not committed to the database. An empty array if no validation is performed.
     # * num_inserts - the number of insert statements it took to import the data
-    # * ids - the primary keys of the imported ids, if the adpater supports it, otherwise and empty array.
+    # * ids - the primary keys of the imported ids if the adapter supports it, otherwise an empty array.
     def import(*args)
       if args.first.is_a?( Array ) && args.first.first.is_a?(ActiveRecord::Base)
         options = {}
@@ -332,11 +408,13 @@ class ActiveRecord::Base
     end
 
     def import_helper( *args )
-      options = { validate: true, timestamps: true, primary_key: primary_key }
+      options = { validate: true, timestamps: true }
       options.merge!( args.pop ) if args.last.is_a? Hash
+      # making sure that current model's primary key is used
+      options[:primary_key] = primary_key
 
       # Don't modify incoming arguments
-      if options[:on_duplicate_key_update]
+      if options[:on_duplicate_key_update] && options[:on_duplicate_key_update].duplicable?
         options[:on_duplicate_key_update] = options[:on_duplicate_key_update].dup
       end
 
@@ -347,58 +425,98 @@ class ActiveRecord::Base
       if args.last.is_a?( Array ) && args.last.first.is_a?(ActiveRecord::Base)
         if args.length == 2
           models = args.last
-          column_names = args.first
+          column_names = args.first.dup
         else
           models = args.first
           column_names = self.column_names.dup
         end
 
+        if models.first.id.nil? && column_names.include?(primary_key) && columns_hash[primary_key].type == :uuid
+          column_names.delete(primary_key)
+        end
+
+        default_values = column_defaults
+        stored_attrs = respond_to?(:stored_attributes) ? stored_attributes : {}
+        serialized_attrs = if defined?(ActiveRecord::Type::Serialized)
+          attrs = column_names.select { |c| type_for_attribute(c.to_s).class == ActiveRecord::Type::Serialized }
+          Hash[attrs.map { |a| [a, nil] }]
+        else
+          serialized_attributes
+        end
+
         array_of_attributes = models.map do |model|
-          # this next line breaks sqlite.so with a segmentation fault
-          # if model.new_record? || options[:on_duplicate_key_update]
           column_names.map do |name|
-            model.read_attribute_before_type_cast(name.to_s)
+            if stored_attrs.key?(name.to_sym) ||
+               serialized_attrs.key?(name) ||
+               default_values.key?(name.to_s)
+              model.read_attribute(name.to_s)
+            else
+              model.read_attribute_before_type_cast(name.to_s)
+            end
+          end
+        end
+        # supports array of hash objects
+      elsif args.last.is_a?( Array ) && args.last.first.is_a?(Hash)
+        if args.length == 2
+          array_of_hashes = args.last
+          column_names = args.first.dup
+        else
+          array_of_hashes = args.first
+          column_names = array_of_hashes.first.keys
+        end
+
+        array_of_attributes = array_of_hashes.map do |h|
+          column_names.map do |key|
+            h[key]
           end
-          # end
         end
         # supports empty array
       elsif args.last.is_a?( Array ) && args.last.empty?
         return ActiveRecord::Import::Result.new([], 0, [])
         # supports 2-element array and array
       elsif args.size == 2 && args.first.is_a?( Array ) && args.last.is_a?( Array )
+
+        unless args.last.first.is_a?(Array)
+          raise ArgumentError, "Last argument should be a two dimensional array '[[]]'. First element in array was a #{args.last.first.class}"
+        end
+
         column_names, array_of_attributes = args
+
+        # dup the passed args so we don't modify unintentionally
+        column_names = column_names.dup
         array_of_attributes = array_of_attributes.map(&:dup)
       else
         raise ArgumentError, "Invalid arguments!"
       end
 
-      # dup the passed in array so we don't modify it unintentionally
-      column_names = column_names.dup
-
       # Force the primary key col into the insert if it's not
       # on the list and we are using a sequence and stuff a nil
       # value for it into each row so the sequencer will fire later
-      if !column_names.include?(primary_key) && connection.prefetch_primary_key? && sequence_name
-        column_names << primary_key
-        array_of_attributes.each { |a| a << nil }
+      symbolized_column_names = Array(column_names).map(&:to_sym)
+      symbolized_primary_key = Array(primary_key).map(&:to_sym)
+
+      if !symbolized_primary_key.to_set.subset?(symbolized_column_names.to_set) && connection.prefetch_primary_key? && sequence_name
+        column_count = column_names.size
+        column_names.concat(primary_key).uniq!
+        columns_added = column_names.size - column_count
+        new_fields = Array.new(columns_added)
+        array_of_attributes.each { |a| a.concat(new_fields) }
       end
 
+      timestamps = {}
+
       # record timestamps unless disabled in ActiveRecord::Base
       if record_timestamps && options.delete( :timestamps )
-        add_special_rails_stamps column_names, array_of_attributes, options
+        timestamps = add_special_rails_stamps column_names, array_of_attributes, options
       end
 
       return_obj = if is_validating
         if models
-          import_with_validations( column_names, array_of_attributes, options ) do |failed|
+          import_with_validations( column_names, array_of_attributes, options ) do |validator, failed|
             models.each_with_index do |model, i|
               model = model.dup if options[:recursive]
-              next if model.valid?(options[:validate_with_context])
-              if options[:raise_error] && model.respond_to?(:raise_validation_error, true) # Rails 5.0 and higher
-                model.send(:raise_validation_error)
-              elsif options[:raise_error] # Rails 3.2, 4.0, 4.1 and 4.2
-                model.send(:raise_record_invalid)
-              end
+              next if validator.valid_model? model
+              raise(ActiveRecord::RecordInvalid, model) if options[:raise_error]
               array_of_attributes[i] = nil
               failed << model
             end
@@ -418,8 +536,8 @@ class ActiveRecord::Base
       return_obj.num_inserts = 0 if return_obj.num_inserts.nil?
 
       # if we have ids, then set the id on the models and mark the models as clean.
-      if support_setting_primary_key_of_imported_objects?
-        set_ids_and_mark_clean(models, return_obj)
+      if models && support_setting_primary_key_of_imported_objects?
+        set_attributes_and_mark_clean(models, return_obj, timestamps)
 
         # if there are auto-save associations on the models we imported that are new, import them as well
         import_associations(models, options.dup) if options[:recursive]
@@ -428,10 +546,6 @@ class ActiveRecord::Base
       return_obj
     end
 
-    # TODO import_from_table needs to be implemented.
-    def import_from_table( options ) # :nodoc:
-    end
-
     # Imports the passed in +column_names+ and +array_of_attributes+
     # given the passed in +options+ Hash with validations. Returns an
     # object with the methods +failed_instances+ and +num_inserts+.
@@ -442,8 +556,10 @@ class ActiveRecord::Base
     def import_with_validations( column_names, array_of_attributes, options = {} )
       failed_instances = []
 
+      validator = ActiveRecord::Import::Validator.new(options)
+
       if block_given?
-        yield failed_instances
+        yield validator, failed_instances
       else
         # create instances for each of our column/value sets
         arr = validations_array_for_column_names_and_attributes( column_names, array_of_attributes )
@@ -453,10 +569,12 @@ class ActiveRecord::Base
         model = new
         arr.each_with_index do |hsh, i|
           hsh.each_pair { |k, v| model[k] = v }
-          next if model.valid?(options[:validate_with_context])
+          next if validator.valid_model? model
           raise(ActiveRecord::RecordInvalid, model) if options[:raise_error]
           array_of_attributes[i] = nil
-          failed_instances << model.dup
+          failure = model.dup
+          failure.errors.send(:initialize_dup, model.errors)
+          failed_instances << failure
         end
       end
 
@@ -502,7 +620,9 @@ class ActiveRecord::Base
       end
 
       columns_sql = "(#{column_names.map { |name| connection.quote_column_name(name) }.join(',')})"
-      insert_sql = "INSERT #{options[:ignore] ? 'IGNORE ' : ''}INTO #{quoted_table_name} #{columns_sql} VALUES "
+      pre_sql_statements = connection.pre_sql_statements( options )
+      insert_sql = ['INSERT', pre_sql_statements, "INTO #{quoted_table_name} #{columns_sql} VALUES "]
+      insert_sql = insert_sql.flatten.join(' ')
       values_sql = values_sql_for_columns_and_attributes(columns, array_of_attributes)
 
       number_inserted = 0
@@ -516,14 +636,17 @@ class ActiveRecord::Base
           # perform the inserts
           result = connection.insert_many( [insert_sql, post_sql_statements].flatten,
             batch_values,
+            options,
             "#{self.class.name} Create Many Without Validations Or Callbacks" )
           number_inserted += result[0]
           ids += result[1]
         end
       else
-        values_sql.each do |values|
-          ids << connection.insert(insert_sql + values)
-          number_inserted += 1
+        transaction(requires_new: true) do
+          values_sql.each do |values|
+            ids << connection.insert(insert_sql + values)
+            number_inserted += 1
+          end
         end
       end
       [number_inserted, ids]
@@ -531,11 +654,10 @@ class ActiveRecord::Base
 
     private
 
-    def set_ids_and_mark_clean(models, import_result)
+    def set_attributes_and_mark_clean(models, import_result, timestamps)
       return if models.nil?
-      import_result.ids.each_with_index do |id, index|
-        model = models[index]
-        model.id = id.to_i
+      models -= import_result.failed_instances
+      models.each do |model|
         if model.respond_to?(:clear_changes_information) # Rails 4.0 and higher
           model.clear_changes_information
         else # Rails 3.2
@@ -543,6 +665,18 @@ class ActiveRecord::Base
         end
         model.instance_variable_set(:@new_record, false)
       end
+
+      # if ids were returned for all models we know all were updated
+      if models.size == import_result.ids.size
+        import_result.ids.each_with_index do |id, index|
+          model = models[index]
+          model.id = id
+
+          timestamps.each do |attr, value|
+            model.send(attr + "=", value)
+          end
+        end
+      end
     end
 
     def import_associations(models, options)
@@ -550,6 +684,7 @@ class ActiveRecord::Base
       # notes:
       #    does not handle associations that reference themselves
       #    should probably take a hash to associations to follow.
+      return if models.nil?
       associated_objects_by_class = {}
       models.each { |model| find_associated_objects_for_import(associated_objects_by_class, model) }
 
@@ -585,7 +720,7 @@ class ActiveRecord::Base
           child.public_send("#{association_reflection.foreign_key}=", model.id)
           # For polymorphic associations
           association_reflection.type.try do |type|
-            child.public_send("#{type}=", model.class.name)
+            child.public_send("#{type}=", model.class.base_class.name)
           end
         end
         associated_objects_by_class[model.class.name][association_reflection.name].concat changed_objects
@@ -599,6 +734,7 @@ class ActiveRecord::Base
       # connection gets called a *lot* in this high intensity loop.
       # Reuse the same one w/in the loop, otherwise it would keep being re-retreived (= lots of time for large imports)
       connection_memo = connection
+
       array_of_attributes.map do |arr|
         my_values = arr.each_with_index.map do |val, j|
           column = columns[j]
@@ -607,11 +743,13 @@ class ActiveRecord::Base
           if val.nil? && column.name == primary_key && !sequence_name.blank?
             connection_memo.next_value_for_sequence(sequence_name)
           elsif column
-            if respond_to?(:type_caster) && type_caster.respond_to?(:type_cast_for_database) # Rails 5.0 and higher
-              connection_memo.quote(type_caster.type_cast_for_database(column.name, val))
-            elsif column.respond_to?(:type_cast_from_user)                                   # Rails 4.2 and higher
+            if respond_to?(:type_caster)                                         # Rails 5.0 and higher
+              type = type_for_attribute(column.name)
+              val = type.type == :boolean ? type.cast(val) : type.serialize(val)
+              connection_memo.quote(val)
+            elsif column.respond_to?(:type_cast_from_user)                       # Rails 4.2
               connection_memo.quote(column.type_cast_from_user(val), column)
-            else                                                                             # Rails 3.2, 4.0 and 4.1
+            else                                                                 # Rails 3.2, 4.0 and 4.1
               if serialized_attributes.include?(column.name)
                 val = serialized_attributes[column.name].dump(val)
               end
@@ -624,35 +762,42 @@ class ActiveRecord::Base
     end
 
     def add_special_rails_stamps( column_names, array_of_attributes, options )
-      AREXT_RAILS_COLUMNS[:create].each_pair do |key, blk|
-        next unless self.column_names.include?(key)
-        value = blk.call
-        index = column_names.index(key) || column_names.index(key.to_sym)
-        if index
-          # replace every instance of the array of attributes with our value
-          array_of_attributes.each { |arr| arr[index] = value if arr[index].nil? }
-        else
-          column_names << key
-          array_of_attributes.each { |arr| arr << value }
+      timestamp_columns = {}
+      timestamps        = {}
+
+      if respond_to?(:all_timestamp_attributes_in_model, true) # Rails 5.1 and higher
+        timestamp_columns[:create] = timestamp_attributes_for_create_in_model
+        timestamp_columns[:update] = timestamp_attributes_for_update_in_model
+      else
+        instance = new
+        timestamp_columns[:create] = instance.send(:timestamp_attributes_for_create_in_model)
+        timestamp_columns[:update] = instance.send(:timestamp_attributes_for_update_in_model)
+      end
+
+      # use tz as set in ActiveRecord::Base
+      timestamp = ActiveRecord::Base.default_timezone == :utc ? Time.now.utc : Time.now
+
+      [:create, :update].each do |action|
+        timestamp_columns[action].each do |column|
+          column = column.to_s
+          timestamps[column] = timestamp
+
+          index = column_names.index(column) || column_names.index(column.to_sym)
+          if index
+            # replace every instance of the array of attributes with our value
+            array_of_attributes.each { |arr| arr[index] = timestamp if arr[index].nil? || action == :update }
+          else
+            column_names << column
+            array_of_attributes.each { |arr| arr << timestamp }
+          end
+
+          if supports_on_duplicate_key_update? && action == :update
+            connection.add_column_for_on_duplicate_key_update(column, options)
+          end
         end
       end
 
-      AREXT_RAILS_COLUMNS[:update].each_pair do |key, blk|
-        next unless self.column_names.include?(key)
-        value = blk.call
-        index = column_names.index(key) || column_names.index(key.to_sym)
-        if index
-          # replace every instance of the array of attributes with our value
-          array_of_attributes.each { |arr| arr[index] = value }
-        else
-          column_names << key
-          array_of_attributes.each { |arr| arr << value }
-        end
-
-        if supports_on_duplicate_key_update?
-          connection.add_column_for_on_duplicate_key_update(key, options)
-        end
-      end
+      timestamps
     end
 
     # Returns an Array of Hashes for the passed in +column_names+ and +array_of_attributes+.
