@@ -8,8 +8,8 @@ module ActiveRecord::Import
 
     def initialize(values, options)
       @values = values
-      @reserved_bytes = options[:reserved_bytes]
-      @max_bytes = options[:max_bytes]
+      @reserved_bytes = options[:reserved_bytes] || 0
+      @max_bytes = options.fetch(:max_bytes) { default_max_bytes }
     end
 
     def parse
@@ -32,7 +32,15 @@ module ActiveRecord::Import
         value_sets << arr if i == (values.size - 1)
       end
 
-      [*value_sets]
+      value_sets
+    end
+
+    private
+
+    def default_max_bytes
+      values_in_bytes = values.sum(&:bytesize)
+      comma_separated_bytes = values.size - 1
+      reserved_bytes + values_in_bytes + comma_separated_bytes
     end
   end
 
