@@ -12,6 +12,11 @@ describe "#import" do
     end
   end
 
+  it "warns you that you're using the library wrong" do
+    error = assert_raise(ArgumentError) { Topic.import %w(title author_name), ['Author #1', 'Book #1', 0] }
+    assert_equal error.message, "Last argument should be a two dimensional array '[[]]'. First element in array was a String"
+  end
+
   it "should not produce an error when importing empty arrays" do
     assert_nothing_raised do
       Topic.import []
@@ -43,6 +48,75 @@ describe "#import" do
           Widget.import Build(3, :widgets)
         end
       end
+
+      context "with uppercase letters" do
+        it "should import models successfully" do
+          assert_difference "Car.count", +3 do
+            Car.import Build(3, :cars)
+          end
+        end
+      end
+    end
+
+    context "that have no primary key" do
+      it "should import models successfully" do
+        assert_difference "Rule.count", +3 do
+          Rule.import Build(3, :rules)
+        end
+      end
+    end
+  end
+
+  describe "with an array of hashes" do
+    let(:columns) { [:title, :author_name] }
+    let(:values) { [{ title: "LDAP", author_name: "Jerry Carter", author_email_address: "jcarter@test.com" }, { title: "Rails Recipes", author_name: "Chad Fowler", author_email_address: "cfowler@test.com" }] }
+
+    it "should import hash data successfully" do
+      assert_difference "Topic.count", +2 do
+        Topic.import values, validate: false
+      end
+    end
+
+    it "should import specified hash data successfully" do
+      assert_difference "Topic.count", +2 do
+        Topic.import columns, values, validate: false
+      end
+
+      Topic.all.each do |t|
+        assert_nil t.author_email_address
+      end
+    end
+  end
+
+  unless ENV["SKIP_COMPOSITE_PK"]
+    describe "with composite primary keys" do
+      it "should import models successfully" do
+        tags = [Tag.new(tag_id: 1, publisher_id: 1, tag: 'Mystery')]
+
+        assert_difference "Tag.count", +1 do
+          Tag.import tags
+        end
+      end
+
+      it "should import array of values successfully" do
+        columns = [:tag_id, :publisher_id, :tag]
+        values = [[1, 1, 'Mystery'], [2, 1, 'Science']]
+
+        assert_difference "Tag.count", +2 do
+          Tag.import columns, values, validate: false
+        end
+      end
+    end
+  end
+
+  describe "with STI models" do
+    it "should import models successfully" do
+      dictionaries = [Dictionary.new(author_name: "Noah Webster", title: "Webster's Dictionary")]
+
+      assert_difference "Dictionary.count", +1 do
+        Dictionary.import dictionaries
+      end
+      assert_equal "Dictionary", Dictionary.first.type
     end
   end
 
@@ -51,6 +125,8 @@ describe "#import" do
     let(:valid_values) { [["LDAP", "Jerry Carter"], ["Rails Recipes", "Chad Fowler"]] }
     let(:valid_values_with_context) { [[1111, "Jerry Carter"], [2222, "Chad Fowler"]] }
     let(:invalid_values) { [["The RSpec Book", ""], ["Agile+UX", ""]] }
+    let(:valid_models) { valid_values.map { |title, author_name| Topic.new(title: title, author_name: author_name) } }
+    let(:invalid_models) { invalid_values.map { |title, author_name| Topic.new(title: title, author_name: author_name) } }
 
     context "with validation checks turned off" do
       it "should import valid data" do
@@ -85,6 +161,22 @@ describe "#import" do
         end
       end
 
+      it "should ignore uniqueness validators" do
+        Topic.import columns, valid_values, validate: true
+        assert_difference "Topic.count", +2 do
+          Topic.import columns, valid_values, validate: true
+        end
+      end
+
+      it "should not alter the callback chain of the model" do
+        attributes = columns.zip(valid_values.first).to_h
+        topic = Topic.new attributes
+        Topic.import [topic], validate: true
+        duplicate_topic = Topic.new attributes
+        Topic.import [duplicate_topic], validate: true
+        assert duplicate_topic.invalid?
+      end
+
       it "should not import invalid data" do
         assert_no_difference "Topic.count" do
           Topic.import columns, invalid_values, validate: true
@@ -100,7 +192,39 @@ describe "#import" do
       it "should report the failed instances" do
         results = Topic.import columns, invalid_values, validate: true
         assert_equal invalid_values.size, results.failed_instances.size
-        results.failed_instances.each { |e| assert_kind_of Topic, e }
+        assert_not_equal results.failed_instances.first, results.failed_instances.last
+        results.failed_instances.each do |e|
+          assert_kind_of Topic, e
+          assert_equal e.errors.count, 1
+        end
+      end
+
+      it "should set ids in valid models if adapter supports setting primary key of imported objects" do
+        if ActiveRecord::Base.support_setting_primary_key_of_imported_objects?
+          Topic.import (invalid_models + valid_models), validate: true
+          assert_nil invalid_models[0].id
+          assert_nil invalid_models[1].id
+          assert_equal valid_models[0].id, Topic.all[0].id
+          assert_equal valid_models[1].id, Topic.all[1].id
+        end
+      end
+
+      it "should set ActiveRecord timestamps in valid models if adapter supports setting primary key of imported objects" do
+        if ActiveRecord::Base.support_setting_primary_key_of_imported_objects?
+          Timecop.freeze(Time.at(0)) do
+            Topic.import (invalid_models + valid_models), validate: true
+          end
+
+          assert_nil invalid_models[0].created_at
+          assert_nil invalid_models[0].updated_at
+          assert_nil invalid_models[1].created_at
+          assert_nil invalid_models[1].updated_at
+
+          assert_equal valid_models[0].created_at, Topic.all[0].created_at
+          assert_equal valid_models[0].updated_at, Topic.all[0].updated_at
+          assert_equal valid_models[1].created_at, Topic.all[1].created_at
+          assert_equal valid_models[1].updated_at, Topic.all[1].updated_at
+        end
       end
 
       it "should import valid data when mixed with invalid data" do
@@ -109,6 +233,30 @@ describe "#import" do
         end
         assert_equal 0, Topic.where(title: invalid_values.map(&:first)).count
       end
+
+      it "should run callbacks" do
+        assert_no_difference "Topic.count" do
+          Topic.import columns, [["invalid", "Jerry Carter"]], validate: true
+        end
+      end
+
+      it "should call validation methods" do
+        assert_no_difference "Topic.count" do
+          Topic.import columns, [["validate_failed", "Jerry Carter"]], validate: true
+        end
+      end
+    end
+  end
+
+  context "without :validation option" do
+    let(:columns) { %w(title author_name) }
+    let(:invalid_values) { [["The RSpec Book", ""], ["Agile+UX", ""]] }
+
+    it "should not import invalid data" do
+      assert_no_difference "Topic.count" do
+        result = Topic.import columns, invalid_values
+        assert_equal 2, result.failed_instances.size
+      end
     end
   end
 
@@ -281,6 +429,12 @@ describe "#import" do
       Topic.import [:id, :author_name, :title], [[99, "Bob Jones", "Topic 99"]]
       assert_equal 99, Topic.last.id
     end
+
+    it "ignores the recursive option" do
+      assert_difference "Topic.count", +1 do
+        Topic.import [:author_name, :title], [["David Chelimsky", "The RSpec Book"]], recursive: true
+      end
+    end
   end
 
   context "ActiveRecord timestamps" do
@@ -365,12 +519,12 @@ describe "#import" do
 
   context "importing through an association scope" do
     { has_many: :chapters, polymorphic: :discounts }.each do |association_type, association|
-      let(:book)   { FactoryGirl.create :book }
-      let(:scope)  { book.public_send association }
-      let(:klass)  { { chapters: Chapter, discounts: Discount }[association] }
-      let(:column) { { chapters: :title,  discounts: :amount  }[association] }
-      let(:val1)   { { chapters: 'A',     discounts: 5        }[association] }
-      let(:val2)   { { chapters: 'B',     discounts: 6        }[association] }
+      book   = FactoryGirl.create :book
+      scope  = book.public_send association
+      klass  = { chapters: Chapter, discounts: Discount }[association]
+      column = { chapters: :title,  discounts: :amount  }[association]
+      val1   = { chapters: 'A',     discounts: 5        }[association]
+      val2   = { chapters: 'B',     discounts: 6        }[association]
 
       context "for #{association_type}" do
         it "works importing models" do
@@ -465,6 +619,30 @@ describe "#import" do
     end
   end
 
+  context 'importing arrays of values with boolean fields' do
+    let(:columns) { [:author_name, :title, :for_sale] }
+
+    it 'should be able to coerce integers as boolean fields' do
+      Book.delete_all if Book.count > 0
+      values = [['Author #1', 'Book #1', 0], ['Author #2', 'Book #2', 1]]
+      assert_difference "Book.count", +2 do
+        Book.import columns, values
+      end
+      assert_equal false, Book.first.for_sale
+      assert_equal true, Book.last.for_sale
+    end
+
+    it 'should be able to coerce strings as boolean fields' do
+      Book.delete_all if Book.count > 0
+      values = [['Author #1', 'Book #1', 'false'], ['Author #2', 'Book #2', 'true']]
+      assert_difference "Book.count", +2 do
+        Book.import columns, values
+      end
+      assert_equal false, Book.first.for_sale
+      assert_equal true, Book.last.for_sale
+    end
+  end
+
   describe "importing when model has default_scope" do
     it "doesn't import the default scope values" do
       assert_difference "Widget.unscoped.count", +2 do
@@ -485,43 +663,100 @@ describe "#import" do
   end
 
   describe "importing serialized fields" do
-    it "imports values for serialized fields" do
+    it "imports values for serialized Hash fields" do
       assert_difference "Widget.unscoped.count", +1 do
         Widget.import [:w_id, :data], [[1, { a: :b }]]
       end
       assert_equal({ a: :b }, Widget.find_by_w_id(1).data)
     end
 
-    if ENV['AR_VERSION'].to_f >= 3.1
-      let(:data) { { a: :b } }
-      it "imports values for serialized JSON fields" do
-        assert_difference "Widget.unscoped.count", +1 do
-          Widget.import [:w_id, :json_data], [[9, data]]
+    it "imports values for serialized fields" do
+      assert_difference "Widget.unscoped.count", +1 do
+        Widget.import [:w_id, :unspecified_data], [[1, { a: :b }]]
+      end
+      assert_equal({ a: :b }, Widget.find_by_w_id(1).unspecified_data)
+    end
+
+    it "imports values for custom coder" do
+      assert_difference "Widget.unscoped.count", +1 do
+        Widget.import [:w_id, :custom_data], [[1, { a: :b }]]
+      end
+      assert_equal({ a: :b }, Widget.find_by_w_id(1).custom_data)
+    end
+
+    let(:data) { { a: :b } }
+    it "imports values for serialized JSON fields" do
+      assert_difference "Widget.unscoped.count", +1 do
+        Widget.import [:w_id, :json_data], [[9, data]]
+      end
+      assert_equal(data.as_json, Widget.find_by_w_id(9).json_data)
+    end
+
+    it "imports serialized values from saved records" do
+      Widget.import [:w_id, :json_data], [[1, data]]
+      assert_equal data.as_json, Widget.last.json_data
+
+      w = Widget.last
+      w.w_id = 2
+      Widget.import([w])
+      assert_equal data.as_json, Widget.last.json_data
+    end
+
+    context "with a store" do
+      it "imports serialized attributes set using accessors" do
+        vendors = [Vendor.new(name: 'Vendor 1', color: 'blue')]
+        assert_difference "Vendor.count", +1 do
+          Vendor.import vendors
         end
-        assert_equal(data.as_json, Widget.find_by_w_id(9).json_data)
+        assert_equal('blue', Vendor.first.color)
       end
     end
   end
 
   describe "#import!" do
-    let(:columns) { %w(title author_name) }
-    let(:valid_values) { [["LDAP", "Jerry Carter"], ["Rails Recipes", "Chad Fowler"]] }
-    let(:invalid_values) { [["Rails Recipes", "Chad Fowler"], ["The RSpec Book", ""], ["Agile+UX", ""]] }
+    context "with an array of unsaved model instances" do
+      let(:topics) { Build(2, :topics) }
+      let(:invalid_topics) { Build(2, :invalid_topics) }
 
-    context "with invalid data" do
-      it "should raise ActiveRecord::RecordInvalid" do
-        assert_no_difference "Topic.count" do
-          assert_raise ActiveRecord::RecordInvalid do
-            Topic.import! columns, invalid_values
+      context "with invalid data" do
+        it "should raise ActiveRecord::RecordInvalid" do
+          assert_no_difference "Topic.count" do
+            assert_raise ActiveRecord::RecordInvalid do
+              Topic.import! invalid_topics
+            end
+          end
+        end
+      end
+
+      context "with valid data" do
+        it "should import data" do
+          assert_difference "Topic.count", +2 do
+            Topic.import! topics
           end
         end
       end
     end
 
-    context "with valid data" do
-      it "should import data" do
-        assert_difference "Topic.count", +2 do
-          Topic.import! columns, valid_values
+    context "with array of columns and array of values" do
+      let(:columns) { %w(title author_name) }
+      let(:valid_values) { [["LDAP", "Jerry Carter"], ["Rails Recipes", "Chad Fowler"]] }
+      let(:invalid_values) { [["Rails Recipes", "Chad Fowler"], ["The RSpec Book", ""], ["Agile+UX", ""]] }
+
+      context "with invalid data" do
+        it "should raise ActiveRecord::RecordInvalid" do
+          assert_no_difference "Topic.count" do
+            assert_raise ActiveRecord::RecordInvalid do
+              Topic.import! columns, invalid_values
+            end
+          end
+        end
+      end
+
+      context "with valid data" do
+        it "should import data" do
+          assert_difference "Topic.count", +2 do
+            Topic.import! columns, valid_values
+          end
         end
       end
     end
