@@ -1,6 +1,10 @@
 class Topic < ActiveRecord::Base
   validates_presence_of :author_name
   validates :title, numericality: { only_integer: true }, on: :context_test
+  validates :title, uniqueness: true
+
+  validate -> { errors.add(:title, :validate_failed) if title == 'validate_failed' }
+  before_validation -> { errors.add(:title, :invalid) if title == 'invalid' }
 
   has_many :books, inverse_of: :topic
   belongs_to :parent, class_name: "Topic"
