@@ -32,6 +32,7 @@ FactoryGirl.define do
   end
 
   factory :rule do
+    sequence(:id) { |n| n }
     sequence(:condition_text) { |n| "q_#{n}_#{n}" }
   end
 
@@ -54,4 +55,9 @@ FactoryGirl.define do
     title 'Tortilla Flat'
     author_name 'John Steinbeck'
   end
+
+  factory :car do
+    sequence(:Name) { |n| n }
+    sequence(:Features) { |n| "Feature #{n}" }
+  end
 end
